unit RememberScrollMain;

interface
uses
  Classes,
  Windows,
  DatulaPlugin;

type
  TRememberScroll = class(TDatulaPlugin)
  private
    m_View: THandle;
    m_List: TStringList;
    m_CurrentMessageID: string;
    m_CurrentPos: integer;
    m_OriginalProc: TFNWndProc;
    m_EmptyCount: integer;

    procedure HookWindowProc( wnd1: THandle );
    procedure UnHookWindowProc;
    procedure RecordIt( message_id: string; pos: integer );
    procedure RestoreIt( message_id: string );
    procedure ForgetIt( message_id: string );

  public
    constructor Create;
    destructor Destroy; override;
    procedure OnUpdateView; override;
    procedure OnActivate(hWnd: THandle; nState: Cardinal; bMinimized: boolean); override;
  end;

implementation
uses
  DatulaApi, SysUtils, Messages;

const
  VERSION_STR = 'test8';
  VERSION_VAL = $1000008;
  PLUGIN_TITLE = 'RememberScrollBar';
  AUTHOR = '';

{ TRememberScroll }

constructor TRememberScroll.Create;
begin
  inherited Create( PLUGIN_TITLE, AUTHOR, VERSION_STR, VERSION_VAL );

  m_List := TStringList.Create;
  m_List.Sorted := true;
  m_List.Duplicates := dupIgnore;

  m_CurrentMessageId := '';
  m_CurrentPos := 0;
  m_EmptyCount := 0;
  m_View := 0;
  m_OriginalProc := nil;
end;

destructor TRememberScroll.Destroy;
begin
  UnHookWindowProc;
  m_List.Free;

  inherited;
end;

procedure TRememberScroll.OnActivate(hWnd: THandle; nState: Cardinal;
  bMinimized: boolean);
var
  window: TMainWindowInfo;
begin
  //Called when the main frame or postman frame are activated.
  //hWnd: The handle of the window which is activated.
  //nState: WA_INACTIVE/WA_ACTIVE/WA_CLICKACTIVE
  //bMinimized: The window is iconified.

  if (nState = WA_INACTIVE) or bMinimized then
    Exit;

  GetMainWindow( window );
  if window.frameWnd = hWnd then
  begin
    HookWindowProc( window.viewer );
  end;
end;

procedure TRememberScroll.RecordIt( message_id: string; pos: integer );
var
  i: integer;
begin
  if pos > 0 then
  begin
    m_List.AddObject( message_id, TObject(pos) );
  end else begin
    if m_List.Find(message_id, i) then
      m_List.Delete(i);
  end;
end;

procedure TRememberScroll.RestoreIt( message_id: string );
var
  i: integer;
begin
  if m_List.Find(message_id, i) then
  begin
    m_CurrentPos := Integer(m_List.Objects[i]);

    SetScrollPos( m_View, SB_VERT, m_CurrentPos, true );
    PostMessage( m_View, WM_VSCROLL,SB_THUMBPOSITION, 0 );
  end;
end;

procedure TRememberScroll.ForgetIt( message_id: string );
begin
  m_CurrentPos := 0;
  SetScrollPos( m_View, SB_VERT, 0, true );
  PostMessage( m_View, WM_VSCROLL,SB_THUMBPOSITION, 0 );

  RecordIt( message_id, 0 );
end;

procedure TRememberScroll.OnUpdateView;
  function GetMessageId( const headertext: string ): string;
  var
    i, j: integer;
    headers: TStringList;
    header_name, header_value: string;
  begin
    result := '';
    headers := TStringList.Create;
    try
      headers.Text := GetHeaderText;
      for i := 0 to headers.Count-1 do
      begin
        j := Pos( ':', headers[i] );
        if j > 0 then begin
          header_name := LowerCase(Copy(headers[i], 1, j - 1));
          header_value := TrimLeft(Copy(headers[i], j + 1, Length(headers[i])));
          headers[i] := header_name + '=' + header_value;
        end;
      end;

      result := headers.Values['message-id'];
    finally
      headers.Free;
    end;
  end;
var
  header_text: string;
  message_id: string;
  last_message_id: string;
begin
  if m_View = 0 then
    Exit;

  header_text := GetHeaderText;

  (* test8ɂJN
    \̃bZ[Wĕ\Ƃ́Awb_擾łȂCxg1񗈂Ƃ
    VbZ[WCxgB
    tH_IAhtml on/offƂȂǂ̓wb_擾łȂCxg2񗈂B
    tHgύX̂Ƃ͈ȂB

    āAĕ\̂Ƃ̂ݖY悤ɂ邽߂ɂ́Ả񐔂1ŁAO
    bZ[WƔfΗǂB
  *)
  //OutputDebugString( Pchar('Remember: header = ' + IntToStr(Length(header_text)) + 'bytes') );

  if header_text = '' then
  begin
    if m_CurrentMessageId <> '' then
    begin
      if m_EmptyCount = 0 then
        RecordIt(m_CurrentMessageId, m_CurrentPos);
      Inc(m_EmptyCount);
      Exit;
    end;
  end;

  message_id := GetMessageId( header_text );
  if message_id = '' then
  begin
    m_EmptyCount := 0;
    Exit;
  end;

  last_message_id := m_CurrentMessageId;

  if message_id <> '' then begin
    if last_message_id = message_id then
      if m_EmptyCount = 1 then
        ForgetIt( message_id );
  end;
  m_EmptyCount := 0;

  if m_CurrentPos > 0 then
    RecordIt( last_message_id, m_CurrentPos );

  m_CurrentMessageId := message_id;
  m_CurrentPos := 0;

  RestoreIt( message_id );
end;

function SubClassProc(hWindow: HWND; Msg: UINT;
                       wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  obj: TRememberScroll;
begin
  obj := TRememberScroll(PluginMain); // asłȂ̂͐\̂ /

  result := CallWindowProc(obj.m_OriginalProc,hWindow,Msg,wParam,lParam);
  if hWindow = obj.m_View then begin
    case Msg of
    WM_COMMAND,
    WM_VSCROLL,
    WM_MOUSEWHEEL,
    WM_KEYDOWN:
      obj.m_CurrentPos := GetScrollPos( hWindow, SB_VERT );
    end;
  end;
end;

procedure TRememberScroll.HookWindowProc( wnd1: THandle );
var
  get_proc: TFNWndProc;
begin
  if  m_View = wnd1 then
    exit;

  UnHookWindowProc;

  m_View := wnd1;

  get_proc := Pointer(GetWindowLong(m_View,GWL_WNDPROC));
  if get_proc <> @SubClassProc then
  begin
    m_OriginalProc := get_proc;
    SetWindowLong(m_View,GWL_WNDPROC,integer(@SubClassProc));
  end;
end;

procedure TRememberScroll.UnHookWindowProc;
begin
  if m_View <> 0 then
  begin
    if Pointer(GetWindowLong(m_View,GWL_WNDPROC)) = @SubClassProc then
    begin
      SetWindowLong(m_View,GWL_WNDPROC,integer(m_OriginalProc));
    end;
    m_View := 0;
  end;
end;

begin
  PluginMain := TRememberScroll.Create;
end.
